import json
import inspect
from datetime import datetime
from types import MappingProxyType
from typing import Any, Union

__all__ = ['JsonObject']


def _attr_is_function(cls, attr):
    return hasattr(cls, attr) and inspect.isfunction(getattr(cls, attr))


class JsonObject:
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            self.__dict__[k] = _load(v)

    def __getitem__(self, key):
        if key in self.__dict__:
            return self.__dict__[key]
        if type(key) is int and key < len(self):
            # assuming insertion order, which should be guaranteed in py 3.6/3.7+ (PEP 520)
            return self.__dict__[list(self.__dict__.keys())[key]]
        return None

    def __setitem__(self, key, value):
        self.__setattr__(key, value)

    def __setattr__(self, key, value):
        if _attr_is_function(self.__class__, key):
            # preserve class methods if an attribute would overwrite it
            # such that if a user sets obj.items = 5, obj.items() will still
            # be callable under the new attribute obj._items(), and obj.items will equal 5
            setattr(self.__class__, f'_{key}', getattr(self.__class__, key))
        self.__dict__[key] = value

    def __getattr__(self, attr):
        # This method is called ONLY when an attribute does NOT exist
        if (
            attr.startswith('_')
            and not attr.startswith('__')
            and _attr_is_function(self.__class__, attr[1:])
        ):
            return getattr(self, attr[1:])

        return None

    def __delitem__(self, key):
        if key in self.__dict__:
            del self.__dict__[key]

    def __contains__(self, key):
        return key in self.__dict__

    def __repr__(self):
        return dumps(self)

    def __str__(self):
        return dumps(self)

    def __iter__(self):
        for k, v in self.__dict__.items():
            yield k, v

    def __len__(self):
        return len(self.__dict__.keys())

    def __delattr__(self, key):
        if key in self.__dict__:
            del self.__dict__[key]

    def __eq__(self, o: object) -> bool:
        # todo: implement equals better
        return str(self) == str(o)

    def keys(self):
        return self.__dict__.keys()

    def keylist(self):
        return list(self.__dict__.keys())

    def items(self):
        return self.__dict__.items()

    def values(self):
        return self.__dict__.values()

    def valuelist(self):
        return list(self.__dict__.values())

    def json(self, indent=None, **kwargs) -> str:
        return dumps(self, indent=indent, **kwargs)

    def dumps(self, indent=None, **kwargs) -> str:
        return dumps(self, indent=indent, **kwargs)

    def pretty(self, indent=2, separators=(', ', ': '), **kwargs) -> str:
        return dumps(self, indent=indent, separators=separators, **kwargs)

    def minify(self, **kwargs):
        return dumps(self, separators=(',', ':'), **kwargs)

    def convert_to(self, cls) -> Any:
        return _load(self, cls=cls)

    def obj(self, cls=dict) -> Union[Any, dict]:
        return self.convert_to(cls)

    def dict(self) -> Union[Any, dict]:
        return self.convert_to(dict)

    def get(self, key: str, default=None) -> Any:
        return self[key] if key in self else default

    def yaml(self, **kwargs) -> str:
        import birdyaml
        return self.convert_to(birdyaml.YamlObject).dumps(**kwargs)

    def write_json(self, filename: str, **kwargs):
        with open(filename, 'w') as f:
            f.write(self.json(**kwargs))

    def write_yaml(self, filename: str, **kwargs):
        with open(filename, 'w') as f:
            f.write(self.yaml(**kwargs))


def load_file(filename: str) -> JsonObject:
    with open(filename) as f:
        return load(f)


def load(fp) -> JsonObject:
    return _load(json.load(fp))


def loads(s: str) -> JsonObject:
    return _load(json.loads(s))


def load_obj(obj: Any, cls=JsonObject) -> JsonObject:
    return _load(obj, cls=cls)


def dumps(obj: Any, indent=None, **kwargs) -> str:
    return json.dumps(obj, indent=indent, default=_json_default_encoder, **kwargs)


def makes(**kwargs) -> str:
    """Create a json string from arguments"""
    return dumps(_load(kwargs, cls=dict))


def make(cls=JsonObject, **kwargs) -> JsonObject:
    """Create a json object from arguments"""
    return _load(kwargs, cls=cls)


def make_obj(cls, **kwargs) -> Any:
    return _load(kwargs, cls=cls)


def make_dict(**kwargs) -> dict:
    return _load(kwargs, cls=dict)


def _load(js, cls: Any = JsonObject) -> Union[JsonObject, Any]:
    if isinstance(js, (list, tuple)):
        return [_load(v, cls=cls) for v in js]
    elif isinstance(js, MappingProxyType):
        return _load_dict(js.copy(), cls=cls)
    elif isinstance(js, dict):
        return _load_dict(js, cls=cls)
    elif hasattr(js, '__dict__'):
        return _load_dict(js.__dict__, cls=cls)
    return js


def _load_dict(d, cls=JsonObject) -> Union[JsonObject, Any]:
    obj = cls()
    for key in d:
        obj[str(key)] = _load(d[key], cls=cls)
    return obj


def _json_default_encoder(o) -> Any:
    if isinstance(o, datetime):
        return str(o)
    elif isinstance(o, MappingProxyType):
        return o.copy()
    elif hasattr(o, '__dict__'):
        return o.__dict__
    return str(o)


if __name__ == '__main__':
    x = make(y=2, json=3)

    # make sure __setitem__ preserves built-in functions when something overwrites it
    print(x.json)
    print(x._json())
    x['minify'] = 10
    print(x.minify)
    print(x._minify())

    # make sure __setattr__ preserves built-in functions when something overwrites it
    assert callable(x.items)
    x.items = [1, 2, 3, 4]
    assert not callable(x.items)
    print(x._items())
    assert callable(x._items)

    assert not callable(x.z)
    x.z = 5
    # Make sure it does not preserve non-callables
    assert x._z is None

    print(x.yaml())
    assert x.yaml() == x._yaml()
